@@ -13,18 +13,36 @@ try:
 except ModuleNotFoundError:
     pypandoc = None
 
+from iotas.attachment_helpers import (
+    AttachmentsCopyOutcome,
+    copy_note_attachments,
+    get_attachment_disk_states,
+    get_attachments_dir,
+)
 from iotas.html_generator import HtmlGenerator
+from iotas.markdown_helpers import (
+    get_image_attachments_from_note_content,
+    get_note_export_content,
+    parse_to_tokens,
+)
 from iotas.note import Note
 from iotas.pdf_exporter import PdfExporter
 
 
 class Exporter(GObject.Object):
+    """Note exporter.
+
+    :param PdfExporter pdf_exporter: PDF exporter
+    :param HtmlGenerator html_generator: HTML generator
+    :param str app_data_path: User data path
+    """
 
     __gsignals__ = {
-        # Out format, out path
-        "finished": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
-        # Out format and reason
-        "failed": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
+        "finished-downloading": (GObject.SignalFlags.RUN_FIRST, None, ()),
+        # str: out path
+        "finished": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
+        # str: reason
+        "failed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
     }
 
     def __init__(
@@ -39,6 +57,7 @@ class Exporter(GObject.Object):
         self.__app_data_path = app_data_path
         self.__active = False
         self.__in_error = False
+        self.__allow_missing_images = False
 
     def export(
         self,
@@ -46,6 +65,7 @@ class Exporter(GObject.Object):
         out_format: str,
         file_extension: str,
         supporting_tex: bool,
+        allow_missing_images: bool = False,
         user_location: Optional[str] = None,
     ) -> None:
         """Export note.
@@ -54,11 +74,14 @@ class Exporter(GObject.Object):
         :param str out_format: Export format
         :param str file_extension: File extension
         :param bool supporting_tex: TeX support
-        :param str user_location: User chosen export location
+        :param bool allow_missing_images: Whether missing images are treated as a failure
+        :param str user_location: User chosen export location, optional
         """
         self.__note = note
         self.__out_format = out_format
         self.__supporting_tex = supporting_tex
+        self.__allow_missing_images = allow_missing_images
+
         if user_location:
             self.__location = user_location
         else:
@@ -73,10 +96,13 @@ class Exporter(GObject.Object):
                 try:
                     os.mkdir(export_dir)
                 except OSError as e:
-                    logging.warning(f"Failed to export {out_format} to {self.__location}: %s", e)
-                    self.emit("failed", self.__out_format, e)
+                    logging.warning(f"Failed to export {out_format} to {self.__location}: {e}")
+                    self.emit("failed", e)
                     return
 
+        if not self.__check_for_missing_attachments():
+            return
+
         if out_format == "pdf":
             self.__export_pdf()
         elif out_format == "md":
@@ -92,7 +118,7 @@ class Exporter(GObject.Object):
     ) -> str:
         """Build an export filename for the note.
 
-        :param Note note: Note to render
+        :param Note note: Note to export
         :param str out_format: Export format
         :param str file_extension: File extension
         :param bool add_timestamp: Whether to prefix timestamp
@@ -103,7 +129,13 @@ class Exporter(GObject.Object):
         if add_timestamp:
             ts = datetime.datetime.now()
             filename = ts.strftime("%Y-%m-%dT%H:%M:%S") + " " + filename
-        if out_format != "html":
+        if out_format == "md":
+            export_to_dir = len(get_image_attachments_from_note_content(note)) > 0
+        elif out_format == "html":
+            export_to_dir = True
+        else:
+            export_to_dir = False
+        if not export_to_dir:
             filename += "." + file_extension
         return filename
 
@@ -117,31 +149,38 @@ class Exporter(GObject.Object):
 
     def __on_pdf_export_finished(self) -> None:
         self.__active = False
-        self.emit("finished", "pdf", self.__location)
+        self.emit("finished", self.__location)
 
     def __on_pdf_export_failed(self, error: str) -> None:
         self.__active = False
-        self.emit("failed", "pdf", error)
+        self.emit("failed", error)
 
     def __export_pandoc(self, out_format: str) -> None:
         self.__active = True
+        content = get_note_export_content(self.__note, prefix_note_id=True)
+        working_dir = os.path.abspath(os.path.join(get_attachments_dir(), os.pardir))
+
         if pypandoc:
             try:
                 pypandoc.convert_text(
-                    self.__note.content, out_format, format="gfm", outputfile=self.__location
+                    content,
+                    out_format,
+                    format="gfm+attributes+implicit_figures",
+                    outputfile=self.__location,
+                    cworkdir=working_dir,
                 )
             except (RuntimeError, OSError) as e:
-                logging.warning(f"Failed to export {out_format} to {self.__location}: %s", e)
-                self.emit("failed", self.__out_format, str(e))
+                logging.warning(f"Failed to export {out_format} to {self.__location}: {e}")
+                self.emit("failed", str(e))
             else:
                 logging.info(f"Exported {self.__out_format} to {self.__location}")
-                self.emit("finished", self.__out_format, self.__location)
+                self.emit("finished", self.__location)
         else:
             logging.warning(
                 f"Failed to export {out_format} to {self.__location}: "
                 "Pypandoc is not installed on your system."
             )
-            self.emit("failed", self.__out_format, "Pypandoc is not installed on your system.")
+            self.emit("failed", "Pypandoc is not installed on your system.")
         self.__active = False
 
     def __export_pdf(self) -> None:
@@ -151,62 +190,71 @@ class Exporter(GObject.Object):
     def __export_md(self) -> None:
         self.__active = True
 
+        if len(get_image_attachments_from_note_content(self.__note)) > 0:
+            file_location = os.path.join(
+                self.__location, f"{self.__sanitise_title_for_filename(self.__note.title)}.md"
+            )
+            if not self.__create_export_directory():
+                return
+        else:
+            file_location = self.__location
+
+        export_content = get_note_export_content(self.__note, prefix_note_id=False)
         try:
-            with open(self.__location, "w") as f:
-                f.write(self.__note.content)
+            with open(file_location, "w") as f:
+                f.write(export_content)
         except OSError as e:
-            self.emit("failed", self.__out_format, e)
-            logging.warning("Failed to export MD: %s", e)
+            self.emit("failed", e)
+            logging.warning(f"Failed to export MD: {e}")
             self.__active = False
             return
 
+        if not self.__copy_attachments():
+            return
+
         self.__active = False
         logging.info(f"Exported {self.__out_format} to {self.__location}")
-        self.emit("finished", self.__out_format, self.__location)
+        self.emit("finished", self.__location)
 
     def __export_html(self) -> None:
         self.__active = True
 
-        (content, _) = self.__html_generator.generate(
-            self.__note, searching=False, export_format="html"
+        parser, tokens = parse_to_tokens(
+            self.__note, exporting=True, tex_support=self.__supporting_tex
+        )
+        html = self.__html_generator.generate(
+            self.__note,
+            tokens,
+            parser.renderer.render,
+            parser.options,
+            searching=False,
+            export_format="html",
         )
 
-        if os.path.exists(self.__location):
-            logging.debug(f"Removing existing {self.__location}")
-            try:
-                shutil.rmtree(self.__location)
-            except OSError as e:
-                self.emit("failed", self.__out_format, e)
-                logging.warning("Failed to replace dir for HTML export: %s", e)
-                self.__active = False
-                return
-
-        try:
-            os.mkdir(self.__location)
-        except OSError as e:
-            self.emit("failed", self.__out_format, e)
-            logging.warning("Failed to make dir for HTML export: %s", e)
-            self.__active = False
+        if not self.__create_export_directory():
             return
 
         index_filename = os.path.join(self.__location, "index.html")
 
         try:
             with open(index_filename, "w") as f:
-                f.write(content)
+                f.write(html)
         except OSError as e:
-            self.emit("failed", self.__out_format, e)
-            logging.warning("Failed to export HTML: %s", e)
+            self.emit("failed", e)
+            logging.warning(f"Failed to export HTML: {e}")
             self.__active = False
             return
 
+        if not self.__copy_attachments():
+            return
+
         css_dir = os.path.join(self.__location, "css")
 
         try:
             os.mkdir(css_dir)
         except OSError as e:
-            self.emit("failed", self.__out_format, e)
-            logging.warning("Failed to export HTML: %s", e)
+            self.emit("failed", e)
+            logging.warning(f"Failed to export HTML: {e}")
             self.__active = False
             return
 
@@ -215,8 +263,8 @@ class Exporter(GObject.Object):
         try:
             shutil.copyfile(f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_CSS_PATH}", dest_file)
         except OSError as e:
-            self.emit("failed", self.__out_format, e)
-            logging.warning("Failed to export HTML: %s", e)
+            self.emit("failed", e)
+            logging.warning(f"Failed to export HTML: {e}")
             self.__active = False
             return
 
@@ -230,8 +278,8 @@ class Exporter(GObject.Object):
                     f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_CSS_PATH}", dest_file
                 )
             except OSError as e:
-                self.emit("failed", self.__out_format, e)
-                logging.warning("Failed to export HTML: %s", e)
+                self.emit("failed", e)
+                logging.warning(f"Failed to export HTML: {e}")
                 self.__active = False
                 return
 
@@ -240,8 +288,8 @@ class Exporter(GObject.Object):
             try:
                 os.mkdir(js_dir)
             except OSError as e:
-                self.emit("failed", self.__out_format, e)
-                logging.warning("Failed to export HTML: %s", e)
+                self.emit("failed", e)
+                logging.warning(f"Failed to export HTML: {e}")
                 self.__active = False
                 return
 
@@ -252,14 +300,14 @@ class Exporter(GObject.Object):
                     f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_JS_PATH}", dest_file
                 )
             except OSError as e:
-                self.emit("failed", self.__out_format, e)
-                logging.warning("Failed to export HTML: %s", e)
+                self.emit("failed", e)
+                logging.warning(f"Failed to export HTML: {e}")
                 self.__active = False
                 return
 
         self.__active = False
         logging.info(f"Exported {self.__out_format} to {self.__location}")
-        self.emit("finished", self.__out_format, self.__location)
+        self.emit("finished", self.__location)
 
     def __sanitise_title_for_filename(self, title: str) -> str:
         """For synced notes the server has already done the sanitising for us. This is for
@@ -267,3 +315,55 @@ class Exporter(GObject.Object):
         """
         value = unicodedata.normalize("NFKC", str(title))
         return re.sub(r"[^\w\s\.-]", "", value).strip()
+
+    def __create_export_directory(self) -> bool:
+        if os.path.exists(self.__location):
+            logging.debug(f"Removing existing {self.__location}")
+            try:
+                shutil.rmtree(self.__location)
+            except OSError as e:
+                self.emit("failed", e)
+                logging.warning(f"Failed to replace dir for {self.__out_format} export: {e}")
+                self.__active = False
+                return False
+
+        try:
+            os.mkdir(self.__location)
+        except OSError as e:
+            self.emit("failed", e)
+            logging.warning(f"Failed to make dir for {self.__out_format} export: {e}")
+            self.__active = False
+            return False
+
+        return True
+
+    def __check_for_missing_attachments(self) -> bool:
+        if len(get_image_attachments_from_note_content(self.__note)) > 0:
+            logging.debug("Checking attachments are on disk")
+            parser, tokens = parse_to_tokens(
+                self.__note, exporting=True, tex_support=self.__supporting_tex
+            )
+            states = get_attachment_disk_states(self.__note, tokens)
+            success = len(states.missing) == 0
+            if not success and not self.__allow_missing_images:
+                self.emit("failed", "Failed to export some attachments")
+                self.__active = False
+                return False
+
+        return True
+
+    def __copy_attachments(self) -> bool:
+        if len(get_image_attachments_from_note_content(self.__note)) > 0:
+            attachments_dir = os.path.join(self.__location, "attachments")
+            logging.debug("Copying attachments")
+            result = copy_note_attachments(self.__note, attachments_dir, prefix_note_id=False)
+            success = result.outcome in (
+                AttachmentsCopyOutcome.NONE,
+                AttachmentsCopyOutcome.SUCCESS,
+            )
+            if not success and not self.__allow_missing_images:
+                self.emit("failed", "Failed to export some attachments")
+                self.__active = False
+                return False
+
+        return True
