@@ -650,7 +650,7 @@ class Index(Adw.BreakpointBin):
         if content is not None and content != "":
             note.content = content
             note.update_title_from_top_line()
-            self.__note_manager.persist_note_while_editing(note)
+            self.__note_manager.persist_note_while_editing(note, update_excerpt=True)
         self.emit("note-opened", note, False)
 
     def __delete_note_from_keyboard(self) -> None:
@@ -966,7 +966,7 @@ class Index(Adw.BreakpointBin):
         :param list[Note] notes: Notes to delete
         """
         self.__sync_manager.flush_pending_deletions()
-        self.__note_manager.delete_notes(notes)
+        self.__note_manager.delete_notes(notes, provide_undo=True)
         self.update_for_note_deletions(notes)
 
     def __is_showing_category(self) -> bool:
