@@ -7,6 +7,7 @@ from threading import Thread
 import time
 from typing import Optional
 
+from iotas.attachment_helpers import delete_attachments_for_note
 from iotas.category import Category
 from iotas.category_manager import CategoryManager
 from iotas.config_manager import ConfigManager
@@ -117,25 +118,36 @@ class NoteManager(GObject.Object):
         self.__category_manager.note_added(note.category)
         return note
 
-    def delete_notes(self, notes: list[Note]) -> None:
+    def delete_notes(self, notes: list[Note], provide_undo: bool) -> None:
         """Delete notes from the database and model.
 
         :param list[Note] notes: The notes
+        :param bool provide_undo: Whether to provide ability to undo
         """
-        self.__deleted_notes = []
+        if provide_undo:
+            for note in self.__deleted_notes:
+                delete_attachments_for_note(note)
+            self.__deleted_notes = []
+
         for note in notes:
-            if not note.content_loaded:
+            if provide_undo and not note.content_loaded:
                 self.__db.populate_note_content(note)
+
             note.locally_deleted = True
-            # If no sync server
-            if note.remote_id >= 0:
+            if note.remote_id >= 0 and provide_undo:
                 self.__db.persist_note_locally_deleted(note, True)
             else:
                 self.__db.delete_note(note.id)
                 self.remove_notes_from_model([note])
-            self.__deleted_notes.append(note)
+
+            if provide_undo:
+                self.__deleted_notes.append(note)
             self.__category_manager.note_deleted(note.category)
 
+            # Delay attachment removal for local deletions, allowing for the undo period
+            if not provide_undo:
+                delete_attachments_for_note(note)
+
     def fetch_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
         """Fetch a note by remote id"
 
@@ -168,11 +180,14 @@ class NoteManager(GObject.Object):
 
     def on_deletion_undo_elapsed(self) -> None:
         """Push local deletions to server and flush local queue upon undo lapse."""
+        had_remote = False
         for note in self.__deleted_notes:
+            delete_attachments_for_note(note)
             if note.has_remote_id:
-                self.emit("sync-requested")
-                break
+                had_remote = True
         self.__deleted_notes = []
+        if had_remote:
+            self.emit("sync-requested")
 
     def undo_deletion(self) -> bool:
         """Undo a pending deletion.
@@ -196,10 +211,21 @@ class NoteManager(GObject.Object):
         self.__deleted_notes = []
         return True
 
-    def persist_note_while_editing(self, note: Note) -> None:
+    def apply_deleted_notes_attachment_deletion(self) -> None:
+        """Delete attachments for any locally deleted notes awaiting undo timeout.
+
+        Ensures the removal of attachments for deleted notes if exiting before the undo toast has
+        lapsed.
+        """
+        for note in self.__deleted_notes:
+            delete_attachments_for_note(note)
+        self.__deleted_notes = []
+
+    def persist_note_while_editing(self, note: Note, update_excerpt: bool) -> None:
         """Persist note to database on separate thread while editing.
 
         :param Note note: The note
+        :param bool update_excerpt: Whether to update excerpt
         """
         if self.__save_timeout_id is None:
             if note.new_and_empty:
@@ -208,12 +234,18 @@ class NoteManager(GObject.Object):
                 self.EDITING_NOTE_SAVE_INTERVAL,
                 self.__persist_note_on_thread,
                 note,
+                update_excerpt,
             )
 
-    def persist_note_sync(self, note: Note) -> None:
+    def persist_note_sync(
+        self, note: Note, allow_empty: bool = False, force_update_excerpt: bool = False
+    ) -> None:
         """Persist note to database on main thread.
 
         :param Note note: The note
+        :param bool allow_empty: Whether to store even when new and empty
+        :param bool force_update_excerpt: Whether to update the excerpt, even if the note isn't
+            "dirty"
         """
         # Cancel any queued save
         if self.__save_timeout_id is not None:
@@ -221,7 +253,8 @@ class NoteManager(GObject.Object):
             self.__save_timeout_id = None
 
         # Persist any changes
-        if note.dirty and not note.handling_conflict and not note.new_and_empty:
+        update = note.dirty or force_update_excerpt
+        if update and not note.handling_conflict and (allow_empty or not note.new_and_empty):
             note.regenerate_excerpt()
             new_note = not note.has_id
             self.__persist_note_to_db(note)
@@ -420,7 +453,7 @@ class NoteManager(GObject.Object):
         try:
             pushes = json.loads(pushes_as_str)
         except json.JSONDecodeError as e:
-            logging.warning("Failed to load failed pushes: %s", e.msg)
+            logging.warning(f"Failed to load failed pushes: {e}")
             return []
 
         if type(pushes) is not list:
@@ -486,9 +519,10 @@ class NoteManager(GObject.Object):
     def search_model(self) -> Optional[NoteListModelSearch]:
         return self.__search_model
 
-    def __persist_note_on_thread(self, note: Note) -> None:
+    def __persist_note_on_thread(self, note: Note, update_excerpt: bool) -> None:
         # Updated here (instead of in Note content setter) to reduce occurrence
-        note.regenerate_excerpt()
+        if update_excerpt:
+            note.regenerate_excerpt()
 
         def remote_sync_and_integrate_new_note(note: Note, is_new: bool) -> None:
             if is_new:
